body{margin:0;font-family:-apple-system,BlinkMacSystemFont,"Segoe UI","Roboto","Oxygen","Ubuntu","Cantarell","Fira Sans","Droid Sans","Helvetica Neue",sans-serif;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}code{font-family:source-code-pro,Menlo,Monaco,Consolas,"Courier New",monospace;margin:0!important}.header{padding:0;background:#001529;height:64px;line-height:64px}.logo{color:#fff;font-size:20px;font-weight:700}.logo,.main{max-width:1200px;margin:0 auto;padding:0 10px}.CodeMirror{resize:vertical;overflow:auto!important}.CodeMirror-foldmarker{font-family:-apple-system,BlinkMacSystemFont,"Segoe UI","Roboto","Oxygen","Ubuntu","Cantarell","Fira Sans","Droid Sans","Helvetica Neue",sans-serif;display:block;color:#999;background-color:#f7f7f7;margin-top:.5em;padding:.5em}.Source-selection{color:#e13c38;background-color:#fdf1f0;border-bottom:1px solid #f2a7a1;border-radius:2px;cursor:help}.Source-tito{color:#72c140;background-color:#f8ffee;border-bottom:1px solid #c1e797;border-radius:2px;cursor:help}.source{border:1px solid #d3d3d3;margin:.2em 0 1em}.source-menu{text-align:right;margin-top:.3em}::-webkit-resizer{border-bottom:2px solid #d3d3d3}.CodeMirror-cursor{display:none!important}.ant-form-item,.ant-form-item .ant-form-item{margin-bottom:0!important}.form-features{margin-top:20px}.form-trace-lengths{margin-top:20px;height:120px}.ant-collapse-arrow{left:0!important}.ant-collapse-header{padding:12px 16px 12px 20px!important}.ant-collapse-content-box{padding:1px!important}.ant-layout-sider{background:none!important}.ant-menu-submenu-title{padding-left:16px!important}.Trace-select .ant-select-selection-item{white-space:normal}.Trace-select .ant-select-selector{overflow:visible;display:block;height:auto!important}
/*# sourceMappingURL=main.d121c257.chunk.css.map */